#include "jabber/jlib.h"

/* Internal routines */
xmlnode _xmlnode_new(pool p, const char* name, unsigned int type)
{
     xmlnode result = NULL;
     if (type > NTYPE_LAST)
	  return NULL;

     if (p == NULL)
          p = pool_new();

     /* Allocate & zero memory */
     result = (xmlnode)pmalloc(p, sizeof(_xmlnode));
     memset(result, '\0', sizeof(_xmlnode));

     /* Initialize fields */
     if (type == NTYPE_CDATA)
	  result->name = pstrdup(p,"#CDATA");
     else
	  result->name = pstrdup(p,name);
     result->type = type;
     result->p = p;
     return result;
}

xmlnode _xmlnode_append_sibling(xmlnode lastsibling, const char* name, unsigned int type)
{
     xmlnode result, current;
     result = _xmlnode_new(lastsibling->p, name, type);
     if (result != NULL)
     {
	  /* Setup sibling pointers */
	  result->prev = lastsibling;
	  lastsibling->next = result;
     }
     return result;
}

xmlnode _xmlnode_insert(xmlnode parent, const char* name, unsigned int type)
{
     xmlnode result;
     /* If parent->firstchild is NULL, simply create a new node for the first child */
     if (parent->firstchild == NULL)
     {
	  result = _xmlnode_new(parent->p, name, type);
	  parent->firstchild = result;
     }
     /* Otherwise, append this to the lastchild */
     else
     {
	  result= _xmlnode_append_sibling(parent->lastchild, name, type);
     }
     result->parent = parent;
     parent->lastchild = result;
     return result;
     
}

xmlnode _xmlnode_search(xmlnode firstsibling, const char* name, unsigned int type)
{
     xmlnode current;

     /* Walk the sibling list, looking for a NTYPE_TAG xmlnode with
	the specified name */
     current = firstsibling;
     while (current != NULL)
     {
	  if ((current->type == type) && (strcmp(current->name, name) == 0))
	       return current;
	  else		    
	       current = current->next;
     }
     return NULL;
}	  

char* _xmlnode_merge(pool p, char* dest, unsigned int destsize, const char* src, unsigned int srcsize)
{
     char* result;
     result = (char*)pmalloc(p, destsize + srcsize + 1);
     memcpy(result, dest, destsize);
     memcpy(result+destsize, src, srcsize);
     result[destsize + srcsize] = '\0';
     return result;
}

/* External routines */
xmlnode xmlnode_new_tag(const char* name)
{
     return _xmlnode_new(NULL, name, NTYPE_TAG);
}

xmlnode xmlnode_insert_tag(xmlnode parent, const char* name)
{
     return _xmlnode_insert(parent, name, NTYPE_TAG);
}

xmlnode xmlnode_insert_cdata(xmlnode parent, const char* CDATA, unsigned int size)
{
     xmlnode result;

     /* If last child is already a CDATA segment, merge this new CDATA segment */
     if ((parent->lastchild != NULL) && (parent->lastchild->type == NTYPE_CDATA))
     {
	  result = parent->lastchild;
	  result->data = _xmlnode_merge(result->p, result->data, result->data_sz, CDATA, size);
	  result->data_sz = result->data_sz + size;
     }
     /* Otherwise, insert a new CDATA node */
     else
     {
	  result = _xmlnode_insert(parent, "", NTYPE_CDATA);
	  if (result != NULL)
	  {
               result->data = (char*)pmalloc(result->p, size + 1);
               memcpy(result->data, CDATA, size);
               result->data[size] = '\0';
	       result->data_sz = size;
	  }
     }
}

xmlnode xmlnode_get_tag(xmlnode parent, const char* name)
{
     if (parent != NULL)
	  return _xmlnode_search(parent->firstchild, name, NTYPE_TAG);

     return NULL;
}	       
     
void xmlnode_put_attrib(xmlnode owner, const char* name, const char* value)
{
     xmlnode attrib;
     /* If there are no existing attributs, allocate a new one to start
	the list */
     if (owner->firstattrib == NULL)
     {
	  attrib = _xmlnode_new(owner->p, name, NTYPE_ATTRIB);
	  owner->firstattrib = attrib;
	  owner->lastattrib  = attrib;
     }
     else
     {
          attrib = _xmlnode_append_sibling(owner->lastattrib, name, NTYPE_ATTRIB);
          owner->lastattrib = attrib;
     }
     /* Update the value of the attribute */
     attrib->data_sz = strlen(value);
     attrib->data    = pstrdup(owner->p, value);

}

char* xmlnode_get_attrib(xmlnode owner, const char* name)
{
     xmlnode attrib;
     
     if (owner->firstattrib != NULL)
     {
	  attrib = _xmlnode_search(owner->firstattrib, name, NTYPE_ATTRIB);
	  if (attrib != NULL)
	       return (char*)attrib->data;
     }
     return NULL;
}

xmlnode xmlnode_get_firstattrib(xmlnode parent)
{
     if (parent != NULL)
	  return parent->firstattrib;
     return NULL;
}

xmlnode xmlnode_get_firstchild(xmlnode parent)
{
     if (parent != NULL)
	  return parent->firstchild;
     return NULL;
}

xmlnode xmlnode_get_lastchild(xmlnode parent)
{
     if (parent != NULL)
	  return parent->firstchild;
     return NULL;
}

xmlnode xmlnode_get_nextsibling(xmlnode sibling)
{
     if (sibling != NULL)
	  return sibling->next;
     return NULL;
}

xmlnode xmlnode_get_prevsibling(xmlnode sibling)
{
     if (sibling != NULL)
	  return sibling->prev;
     return NULL;
}

xmlnode xmlnode_get_parent(xmlnode node)
{
     if (node != NULL)
	  return node->parent;
     return NULL;
}

xmlnode xmlnode_get_child(xmlnode parent, const char* name)
{
     xmlnode child;
     child = NULL;
     if (parent->firstchild != NULL)
     {
          child = _xmlnode_search(parent->firstchild, name, NTYPE_TAG);
     }
     return child;
}

char* xmlnode_get_name(xmlnode node)
{
     if (node != NULL)
	  return node->name;
     return NULL;
}

char* xmlnode_get_data(xmlnode node)
{
     if (node != NULL)
	  return node->data;
     return NULL;
}

int xmlnode_get_datasz(xmlnode node)
{
     if (node != NULL)
	  return node->data_sz;
     return (int)NULL;
}

int xmlnode_get_type(xmlnode node)
{
     if (node != NULL)
	  return node->type;
     return (int)NULL;
}

int xmlnode_has_children(xmlnode node)
{
     if ((node != NULL) && (node->firstchild != NULL))
	  return 1;
     return 0;
}

int xmlnode_has_attribs(xmlnode node)
{
     if ((node != NULL) && (node->firstattrib != NULL))
	  return 1;
     return 0;
}

pool xmlnode_pool(xmlnode node)
{
    if (node != NULL)
        return node->p;
    return (pool)NULL;
}
			    
char *xmlnode2str(xmlnode node)
{
    char *ret, *tmp;
    int len;
    static int refcount = 0;

    if(node == NULL)
        return;

    refcount++;

    switch(xmlnode_get_type(node))
    {
        case NTYPE_ATTRIB:
            ret = strgrow(NULL," ",0,0);
            ret = strgrow(ret,xmlnode_get_name(node),1,0);
            ret = strgrow(ret,"='",1,0);
            ret = strgrow(ret,xmlnode_get_data(node),1,0);
            ret = strgrow(ret,"'",1,0);
            break;
        case NTYPE_TAG:
            ret = strgrow(NULL,"<",0,0);
            ret = strgrow(ret,xmlnode_get_name(node),1,0);
            if (xmlnode_has_attribs(node))
                ret = strgrow(ret,xmlnode2str(xmlnode_get_firstattrib(node)),1,1);
            if (xmlnode_has_children(node))
            {
                ret = strgrow(ret,">",1,0);
                ret = strgrow(ret,xmlnode2str(xmlnode_get_firstchild(node)),1,1);
                ret = strgrow(ret,"</",1,0);
                ret = strgrow(ret,xmlnode_get_name(node),1,0);
                ret = strgrow(ret,">",1,0);
            }else{
                ret = strgrow(ret,"/>",1,0);
            }
            break;
        case NTYPE_CDATA:
            ret = strgrow(NULL,xmlnode_get_data(node),0,0);
            break;
    }

    if(refcount == 1)
    { /* this is the first call, not part of the recursion */
        /* return memory from the pool, free the memory used to build the result */
        refcount = 0;
        tmp = ret;
        ret = pstrdup(node->p, ret);
        free(tmp);
        return ret;
    }else{
        refcount--;
        return strgrow(ret,xmlnode2str(xmlnode_get_nextsibling(node)),1,1);
    }

}     

int xmlnode_cmp(xmlnode a, xmlnode b)
{
    /* !!! needs to be written, loop through both a and b comparing everything, attribs, cdata, children, etc */

    return 0;
}

xmlnode xmlnode_dup(xmlnode x)
{
    /* !!! needs to be written, produce full duplicate of x */
    return x;
}

void xmlnode_free(xmlnode node)
{
    if(node == NULL)
        return;

    pool_free(node->p);
}